@@ -50,6 +50,8 @@ enum AskType {
 	ASK_KEYRING_PASSWORD,
 	ASK_ITEM_READ_WRITE_ACCESS,
 	ASK_NEW_KEYRING_PASSWORD,
+	ASK_ORIGINAL_CHANGE_KEYRING_PASSWORD,
+	ASK_CHANGE_KEYRING_PASSWORD,
 	ASK_DEFAULT_KEYRING
 };
 
@@ -119,23 +121,23 @@ static void
 init_salt (guchar salt[8])
 {
 	gboolean got_random;
-	int i;
+	int i, fd;
 
 	got_random = FALSE;
-#ifdef __linux__
-	{
-		int fd;
-
-		fd = open ("/dev/random", O_RDONLY);
-		if (fd != -1) {
+#ifdef HAVE_DEVRANDOM
+	fd = open ("/dev/random", O_RDONLY);
+	if (fd != -1) {
+		struct stat st;
+		/* Make sure it's a character device */
+		if ((fstat (fd, &st) == 0) && S_ISCHR (st.st_mode)) {
 			if (read (fd, salt, 8) == 8) {
 				got_random = TRUE;
 			}
-			close (fd);
 		}
-		
+		close (fd);			
 	}
 #endif
+
 	if (!got_random) {
 		for (i=0; i < 8; i++) {
 			salt[i] = (int) (256.0*rand()/(RAND_MAX+1.0));
@@ -337,22 +339,27 @@ gnome_keyring_application_ref_new_from_pid (pid_t pid)
 
 	app_ref = g_new0 (GnomeKeyringApplicationRef, 1);
 
-#ifdef __linux__
+#if defined(__linux__) || defined(__FreeBSD__)
 	{
-		char buffer[1024];
+		char *buffer;
 		int len;
-		char *path;
+		char *path = NULL;
 		
-		path = g_strdup_printf ("/proc/%d/exe", pid); 
-		len = readlink (path, buffer, sizeof (buffer));
+#if defined(__linux__)
+		path = g_strdup_printf ("/proc/%d/exe", pid);
+#elif defined(__FreeBSD__)
+		path = g_strdup_printf ("/proc/%d/file", pid);
+#endif
+		buffer = g_file_read_link (path, NULL);
 		g_free (path);
 
+		len = (buffer != NULL) ? strlen (buffer) : 0;
 		if (len > 0) {
 			app_ref->pathname = g_malloc (len + 1);
 			memcpy (app_ref->pathname, buffer, len);
 			app_ref->pathname[len] = 0;
 		}
-		
+		g_free (buffer);
 	}
 #endif
 
@@ -564,6 +571,21 @@ access_request_for_new_keyring_password (const char *keyring_name)
 	return access_request;
 }
 
+static GnomeKeyringAccessRequest *
+access_request_for_change_keyring_password (const char *keyring_name,
+					    gboolean need_original)
+{
+	GnomeKeyringAccessRequest *access_request;
+	access_request = g_new0 (GnomeKeyringAccessRequest, 1);
+	if (need_original) {
+		access_request->request_type = GNOME_KEYRING_ACCESS_REQUEST_ORIGINAL_CHANGE_KEYRING_PASSWORD;
+	} else {
+		access_request->request_type = GNOME_KEYRING_ACCESS_REQUEST_CHANGE_KEYRING_PASSWORD;
+	}
+	access_request->new_keyring = g_strdup (keyring_name);
+	return access_request;
+}
+
 static GnomeKeyringAccessRequest *
 access_request_default_keyring (void)
 {
@@ -613,6 +635,19 @@ find_item (GnomeKeyring *keyring, guint32 id)
 	return find_item_in_list (keyring->items, id);
 }
 
+
+static GnomeKeyringResult
+change_keyring_password (GnomeKeyring *keyring,  const char *password)
+{
+	if (keyring->locked) {
+		return GNOME_KEYRING_RESULT_DENIED;
+	} else { 
+		keyring->password = g_strdup (password);
+		save_keyring_to_disk (keyring);
+		return GNOME_KEYRING_RESULT_OK;
+	}
+}
+
 static GnomeKeyringResult
 unlock_keyring (GnomeKeyring *keyring, const char *password)
 {
@@ -925,6 +960,7 @@ create_new_keyring (const char *keyring_name, const char *password)
 
 }
 
+
 static gboolean
 op_create_keyring_execute (GString *packet,
 			   GString *result,
@@ -1050,6 +1086,130 @@ op_delete_keyring_execute (GString *packet,
 	return TRUE;
 }
 
+static gboolean
+op_change_keyring_password_collect (GString *packet,
+				    GList **access_requests_out)
+{
+	GList *access_requests;
+	GnomeKeyringOpCode opcode;
+	char *keyring_name, *original, *password;
+	GnomeKeyring *keyring;
+	
+	if (!gnome_keyring_proto_decode_op_string_string_string (packet,
+							  &opcode,
+							  &keyring_name,
+							  &original,
+							  &password)) {
+		return FALSE;
+	}
+
+	access_requests = NULL;
+	
+	if (keyring_name == NULL) {
+		keyring_name = "default";
+	}
+	
+	keyring = find_keyring (keyring_name);
+	if (keyring == NULL) {
+		/* don't exist */
+		return FALSE;
+	}
+	
+	if (password == NULL) {
+		if (original == NULL ) {
+			/* Prompt for original and Let user pick password */
+			access_requests =
+				g_list_prepend (access_requests,
+						access_request_for_change_keyring_password (keyring_name, TRUE));
+		} else {
+			/* Use original given to us Let user pick password */
+			access_requests =
+				g_list_prepend (access_requests,
+						access_request_for_change_keyring_password (keyring_name, FALSE));
+		}
+	}
+
+	*access_requests_out = access_requests;
+	
+	g_free (keyring_name);
+	gnome_keyring_free_password (original);
+	gnome_keyring_free_password (password);
+	
+	return TRUE;
+}
+
+static gboolean
+op_change_keyring_password_execute (GString *packet,
+			   GString *result,
+			   GnomeKeyringApplicationRef *app_ref,
+			   GList *access_requests)
+{
+	char *keyring_name, *original, *password;
+	GnomeKeyring *keyring;
+	GnomeKeyringOpCode opcode;
+	GnomeKeyringAccessRequest *req;
+	
+	if (!gnome_keyring_proto_decode_op_string_string_string (packet,
+							  &opcode,
+							  &keyring_name,
+							  &original,
+							  &password)) {
+		return FALSE;
+	}
+	g_assert (opcode == GNOME_KEYRING_OP_CHANGE_KEYRING_PASSWORD);
+	
+	if (keyring_name == NULL) {
+		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_BAD_ARGUMENTS);
+		goto out;
+	}
+	
+	keyring = find_keyring (keyring_name);
+	
+	if (keyring == NULL) {
+		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
+		goto out;
+	}
+	
+	if (original == NULL) {
+		if (access_requests != NULL) {
+			req = access_requests->data;
+			original = g_strdup (req->original);
+		}
+	}
+
+	if (original ==NULL) {
+		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_BAD_ARGUMENTS);
+		goto out;
+	}
+	
+	lock_keyring(keyring);
+	
+	if ( unlock_keyring(keyring, original) != GNOME_KEYRING_RESULT_OK ) {
+		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
+		goto out;
+	}
+	
+	if (password == NULL) {
+		if (access_requests != NULL) {
+			req = access_requests->data;
+			password = g_strdup (req->password);
+		}
+	}
+	
+	if (password == NULL) {
+		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_BAD_ARGUMENTS);
+		goto out;
+	}
+	
+	gnome_keyring_proto_add_uint32 (result, change_keyring_password (keyring, password));
+	
+ out:
+	g_free (keyring_name);
+	gnome_keyring_free_password (original);
+	gnome_keyring_free_password (password);
+	
+	return TRUE;
+}
 
 static gboolean
 op_list_items_collect (GString *packet,
@@ -1150,7 +1310,8 @@ op_create_item_collect (GString *packet,
 	
 	if (!gnome_keyring_proto_decode_create_item (packet,
 						     &keyring_name, NULL,
-						     &attributes, NULL, &type,
+						     &attributes, NULL,
+						     (GnomeKeyringItemType *) &type,
 						     &update_if_exists)) {
 		return FALSE;
 	}
@@ -1230,7 +1391,7 @@ op_create_item_execute (GString *packet,
 						     &display_name,
 						     &attributes,
 						     &secret,
-						     &type,
+						     (GnomeKeyringItemType *) &type,
 						     &update_if_exists)) {
 		return FALSE;
 	}
@@ -1389,8 +1550,8 @@ op_delete_item_execute (GString *packet,
 
 	access_request = access_requests->data;
 
-	if (access_request->item->keyring != keyring ||
-	    access_request->item == NULL ||
+	if (access_request->item == NULL ||
+	    access_request->item->keyring != keyring ||
 	    access_request->item->locked) {
 		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
 		goto out;
@@ -1748,7 +1909,7 @@ op_set_item_info_execute (GString *packet,
 	if (!gnome_keyring_proto_decode_set_item_info (packet,
 						       &keyring_name,
 						       &item_id,
-						       &type,
+						       (GnomeKeyringItemType *) &type,
 						       &item_name,
 						       &secret)) {
 		return FALSE;
@@ -1772,8 +1933,8 @@ op_set_item_info_execute (GString *packet,
 
 	access_request = access_requests->data;
 
-	if (access_request->item->keyring != keyring ||
-	    access_request->item == NULL ||
+	if (access_request->item == NULL ||
+	    access_request->item->keyring != keyring ||
 	    access_request->item->locked) {
 		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
 		goto out;
@@ -1802,6 +1963,41 @@ out:
 	return TRUE;
 }
 
+static gboolean
+op_set_daemon_display_execute (GString *packet,
+			       GString *result,
+			       GnomeKeyringApplicationRef *app_ref,
+			       GList *access_requests)
+{
+       char *display;
+       GnomeKeyringOpCode opcode;
+
+       if (!gnome_keyring_proto_decode_op_string (packet,
+						  &opcode,
+						  &display)) {
+               return FALSE;
+       }
+
+       if ( display == NULL ) {
+               gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
+               goto out;
+       }
+
+       if (!have_display && (g_strrstr (display, ":") != NULL)) {
+               g_setenv ( "DISPLAY", display, TRUE );
+               have_display = TRUE;
+       } else {
+               gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
+               goto out;
+       }
+
+       gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_OK);
+
+out:
+    g_free (display);
+       return TRUE;
+}
+
 static int
 unmatched_attributes (GnomeKeyringAttributeList *attributes,
 		      GnomeKeyringAttributeList *matching)
@@ -1874,8 +2070,8 @@ op_set_item_attributes_execute (GString *packet,
 
 	access_request = access_requests->data;
 
-	if (access_request->item->keyring != keyring ||
-	    access_request->item == NULL ||
+	if (access_request->item == NULL ||
+	    access_request->item->keyring != keyring ||
 	    access_request->item->locked) {
 		gnome_keyring_proto_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
 		goto out;
@@ -2033,6 +2229,7 @@ finish_ask_io (GnomeKeyringAsk *ask,
 	gchar **lines;
 	int response;
 	char *str;
+	char *str1;
 	GnomeKeyring *keyring;
 	GnomeKeyringItem *item;
 
@@ -2042,11 +2239,19 @@ finish_ask_io (GnomeKeyringAsk *ask,
 	/* default for failed requests */
 	response = GNOME_KEYRING_ASK_RESPONSE_FAILURE;
 	str = NULL;
+	str1 = NULL;
 
 	if (!failed) {
 		lines = g_strsplit (ask->buffer->str, "\n", 3);
-		response = atol (lines[0]);
-		str = g_strdup (lines[1]);
+		if (lines[0]) {
+			response = atol (lines[0]);
+			if (lines[1]) {
+				str = g_strdup (lines[1]);
+				if (lines[2]) {
+					str1 = g_strdup (lines[2]);
+				}
+			} 
+		}
 		g_strfreev (lines);
 	}
 
@@ -2055,10 +2260,9 @@ finish_ask_io (GnomeKeyringAsk *ask,
 			ask->access_requests = g_list_remove (ask->access_requests, ask->current_request);
 			gnome_keyring_access_request_free (ask->current_request);
 	} else {
-		
 		switch (ask->current_ask_type) {
 		case ASK_DEFAULT_KEYRING:
-			if (strlen (str) > 0) {
+			if (str && strlen (str) > 0) {
 				keyring = create_new_keyring ("default", str);
 				if (keyring != NULL) {
 					default_keyring = keyring;
@@ -2068,7 +2272,20 @@ finish_ask_io (GnomeKeyringAsk *ask,
 			}
 			break;
 		case ASK_NEW_KEYRING_PASSWORD:
-			if (strlen (str) > 0) {
+			if (str && strlen (str) > 0) {
+				ask->current_request->password = g_strdup (str);
+				ask->current_request->granted = TRUE;
+			}
+			break;
+		case ASK_ORIGINAL_CHANGE_KEYRING_PASSWORD:
+			if (str && strlen (str) > 0 && str1 && strlen (str1) > 0) {
+				ask->current_request->original = g_strdup (str);
+				ask->current_request->password = g_strdup (str1);
+				ask->current_request->granted = TRUE;
+			}
+			break;
+		case ASK_CHANGE_KEYRING_PASSWORD:
+			if (str && strlen (str) > 0) {
 				ask->current_request->password = g_strdup (str);
 				ask->current_request->granted = TRUE;
 			}
@@ -2089,12 +2306,14 @@ finish_ask_io (GnomeKeyringAsk *ask,
 			break;
 		case ASK_ITEM_READ_WRITE_ACCESS:
 			item = ask->current_request->item;
+			keyring = ask->current_request->item->keyring;
 			g_assert (item != NULL);
 			if (response == GNOME_KEYRING_ASK_RESPONSE_ALLOW_FOREVER) {
 				add_item_acl (item, ask->app_ref,
 					      GNOME_KEYRING_ACCESS_READ |
 					      GNOME_KEYRING_ACCESS_WRITE |
 					      GNOME_KEYRING_ACCESS_REMOVE);
+				save_keyring_to_disk (keyring);
 			}
 			ask->current_request->granted = TRUE;
 			/* ok */
@@ -2193,6 +2412,10 @@ launch_ask_helper (GnomeKeyringAsk *ask,
 		argv[1] = "-k";
 	} else if (ask_type == ASK_NEW_KEYRING_PASSWORD) {
 		argv[1] = "-n";
+	} else if (ask_type == ASK_CHANGE_KEYRING_PASSWORD) {
+		argv[1] = "-c";
+	} else if (ask_type == ASK_ORIGINAL_CHANGE_KEYRING_PASSWORD) {
+		argv[1] = "-o";
 	} else if (ask_type == ASK_ITEM_READ_WRITE_ACCESS) {
 		argv[1] = "-i";
 	} else if (ask_type == ASK_DEFAULT_KEYRING) {
@@ -2270,6 +2493,24 @@ schedule_ask (GnomeKeyringAsk *ask)
 			/* iterate */
 			gnome_keyring_ask_iterate (ask);
 		}
+		break;	
+	case GNOME_KEYRING_ACCESS_REQUEST_ORIGINAL_CHANGE_KEYRING_PASSWORD:
+		if (!launch_ask_helper (ask, ASK_ORIGINAL_CHANGE_KEYRING_PASSWORD)) {
+			/* no way to allow request, denying */
+			ask->access_requests = g_list_remove (ask->access_requests, request);
+			gnome_keyring_access_request_free (request);
+			/* iterate */
+			gnome_keyring_ask_iterate (ask);
+		}
+		break;
+	case GNOME_KEYRING_ACCESS_REQUEST_CHANGE_KEYRING_PASSWORD:
+		if (!launch_ask_helper (ask, ASK_CHANGE_KEYRING_PASSWORD)) {
+			/* no way to allow request, denying */
+			ask->access_requests = g_list_remove (ask->access_requests, request);
+			gnome_keyring_access_request_free (request);
+			/* iterate */
+			gnome_keyring_ask_iterate (ask);
+		}
 		break;
 	case GNOME_KEYRING_ACCESS_REQUEST_DEFAULT_KEYRING:
 		if (!launch_ask_helper (ask, ASK_DEFAULT_KEYRING)) {
@@ -2431,6 +2672,8 @@ GnomeKeyringOperationImplementation keyring_ops[] = {
 	{ op_set_item_info_or_attributes_collect, op_set_item_attributes_execute}, /* SET_ITEM_ATTRIBUTES */
 	{ op_get_item_info_or_attributes_collect, op_get_item_acl_execute}, /* GET_ITEM_ACL */
 	{ op_set_item_info_or_attributes_collect, op_set_item_acl_execute}, /* SET_ITEM_ACL */
+	{ op_change_keyring_password_collect, op_change_keyring_password_execute }, /*CHANGE_KEYRING_PASSWORD*/
+ 	{ NULL, op_set_daemon_display_execute}, /* SET_DAEMON_DISPLAY */
 };
 
 static RETSIGTYPE
@@ -2459,12 +2702,10 @@ close_stdinout (void)
 	int fd;
 	
 	fd = open ("/dev/null", O_RDONLY);
-	close (0);
 	sane_dup2 (fd, 0);
 	close (fd);
 	
 	fd = open ("/dev/null", O_WRONLY);
-	close (1);
 	sane_dup2 (fd, 1);
 	close (fd);
 }
